/* $Id: ledpwm.c,v 1.3 2010/07/10 07:34:51 simimeie Exp $
 * Functions for led brightness control via PWM (pulse width modulation).
 */

#include <avr/io.h>
#include "ledpwm.h"

/* Select between Phase correct PWM and Fast PWM mode.
 * setting this to 1 selects phase correct mode, everything else is Fast PWM. */
#define PHASECORRECTPWM 1

/*
 * Our hardware connections are as follows:
 *  Red -> PD6 / OC0A
 *  Green -> PD5 / OC0B
 *  Blue -> PD3 / OC2B
 */

uint8_t ledpwm_re = 0xff, ledpwm_gr = 0xff, ledpwm_bl = 0xff;
uint8_t ledpwm_bri = 128;

void ledpwm_init(void)
{
  /* Set OC2B, OC0A and OC0B to output */
  DDRD |= _BV(3) | _BV(5) | _BV(6);
  /* Set compare output mode for OC2B in timer counter control register: */
#if (PHASECORRECTPWM == 1)
  /* select phase correct PWM mode (1). WGM2 bit is 0 so we don't need to touch TCCR2B. */
  TCCR2A |= _BV(COM2B1) | _BV(WGM20);
#else  /* not PHASECORRECTPWM */
  /* set output at BOTTOM (=0), clear it on compare match.
   * select fast PWM mode 3. WGM2 bit is 0 so we don't need to touch TCCR2B. */
  TCCR2A |= _BV(COM2B1) | _BV(WGM20) /*| _BV(WGM21)*/;
#endif /* PHASECORRECTPWM */
  /* select clock source: cpu clock without prescaler. */
  TCCR2B |= _BV(CS20);
  /* Default brightness: PWM 0 */
  OCR2B = 0;
  /* Set compare output mode for OC0A and OC0B in timer counter control register: */
#if (PHASECORRECTPWM == 1)
  /* select phase correct PWM mode (1). WGM2 bit is 0 so we don't need to touch TCCR0B. */
  TCCR0A |= _BV(COM0A1) | _BV(COM0B1) | _BV(WGM00);
#else  /* not PHASECORRECTPWM */
  /* set output at BOTTOM (=0), clear it on compare match.
   * select fast PWM mode 3. WGM2 bit is 0 so we don't need to touch TCCR0B. */
  TCCR0A |= _BV(COM0A1) | _BV(COM0B1) | _BV(WGM00) | _BV(WGM01);
#endif /* PHASECORRECTPWM */
  /* select clock source: i/o clock without prescaler. */
  TCCR0B |= _BV(CS20);
  /* Default brightness: PWM 0 */
  OCR0A = 0;
  OCR0B = 0;
}

/* In none-phase-correct PWM mode, This function does some mapping of
 * the values:
 * 0 really turns the LED off.
 * 1 is an PWM value of 0, which still lets the LED glimmer.
 * 2 to 254 are mapped to PWM 3 to 253 (always one less!)
 * 255 is mapped to PWM 255.
 * So there is no way to set PWM 254, but you cannot visually see the
 * difference to 253 or 255 anyways...
 * In phase-correct PWM mode there does not need to be a mapping, 0 means off.
 */
void ledpwm_setled(uint8_t led, uint8_t val)
{
#if (PHASECORRECTPWM == 1)
  switch (led) {
  case LEDPWM_REDLED:   OCR0A = val; break;
  case LEDPWM_GREENLED: OCR0B = val; break;
  case LEDPWM_BLUELED:  OCR2B = val; break;
  };
#else /* not PHASECORRECTPWM */
  if (val == 0) {
    switch (led) {
    case LEDPWM_REDLED:   DDRD &= (uint8_t)~_BV(6);
                          break;
    case LEDPWM_GREENLED: DDRD &= (uint8_t)~_BV(5);
                          break;
    case LEDPWM_BLUELED:  DDRD &= (uint8_t)~_BV(3);
                          break;
    };
  } else {
    val -= 1;
    if (val == 254) {
      val = 255;
    }
    switch (led) {
    case LEDPWM_REDLED:   DDRD |= _BV(6);
                          OCR0A = val;
                          break;
    case LEDPWM_GREENLED: DDRD |= _BV(5);
                          OCR0B = val;
                          break;
    case LEDPWM_BLUELED:  DDRD |= _BV(3);
                          OCR2B = val;
                          break;
    };
  }
#endif /* PHASECORRECTPWM */
}

void ledpwm_set(uint8_t red, uint8_t green, uint8_t blue, uint8_t br) {
  ledpwm_setled(LEDPWM_REDLED,  (((uint16_t)red   * br) / 255));
  ledpwm_setled(LEDPWM_GREENLED,(((uint16_t)green * br) / 255));
  ledpwm_setled(LEDPWM_BLUELED, (((uint16_t)blue  * br) / 255));
}
